import numpy as np

FUNCS = {
    'mean' : lambda c, co, *args : c[c>co].mean(),
    'sum'  : lambda c, co, *args : c[c>co].sum(),
    'max'  : lambda c, co, *args : c[c>co].max(),
    'lnr'  : lambda c, co, t, *args : c[c>co].sum() / (t.time[-1] - t.time[0]),
    'nlnr' : lambda c, co, t, cl, *args : c[c>co].sum() / (cl * (t.time[-1] - t.time[0])),
    'occ'  : lambda c, co, t, *args :  100 * (c.sum() / t.n_frames)
}

def calculate_metric(p, func, co=0, t=None, norm=False, unit='us', *args):
    """Calculate default ProLint metrics. The function is designed such
    that it can be easily extended for other custom metrics. The second argument
    has to be a callable object, which is called using the following
    default arguments: contacts, cutoff, MDTraj.Trajectory, number_of_contacts, *args.

    By default ProLint offers the following metrics:

    ``mean`` (Mean_Duration) : The average duration of all contacts.

    ``max`` (Longest_Duration) : The longest duration (averaged if more than 1 protein).

    ``sum`` (Sum_of_all_Contacts) : The total sum of all contacts.

    ``lnr`` (Lipid_Number) : The average number of lipids in contact (the total number of contacts
    normalized with respect to the number of frames).

    ``nlnr`` (Normalized_Lipid_Number) : Lipid_Number normalized with respect to the
    number of different lipids (e.g. number of different cholesterols).

    ``occ`` (Occupancy) : For each frame, we give a value of 0 if no lipid of interest is in
    contact with the residue and 1 otherwise. Occupancy is then: sum_of_all_1s / nr_of_frames.


    Parameters
    ----------

    p : Per residue contacts.
        This is the output of ProLint.retrieve_contacts() call.

    func : callable object.
        Custom function or one of the default ones: FUNC[x] where x is
        one of the following: 'mean', 'sum', 'max', 'lnr', 'nlnr', 'occ'. If you
        want to call FUNCS['occ'] make sure to specify the argument contact='occupancy' to
        the retrieve_contacts call first.

    co : int.
        Discard contacts shorter than co value. Useful when you want to discard contacts
            with a short duration

    t : MDTraj.Trajectory

    norm : bool, default=False,
        Normalize contacts with respect to time

    unit : str
        Time unit to use for the normalization. Either 'us' or 'ns'.

    *args : additional arguments.
        These arguments are passed on to the callable object: func.

    Returns
    -------
    contacts : dict
        Dictionary where the keys are lipids and values are tuples containing
        the mean and standard deviation of the calcualted metric.


    """

    per_replicate = {}
    for lipid, v in p.items():
        r = [func(c, co, t, len(v), *args) if len(c[c>co]) > 0 else 0 for c in v]
        if norm:
            r = [x / 1000. if unit == 'ns' else x / 1000000. for x in r]
        per_replicate[lipid] = (np.mean(r), np.std(r))

    return per_replicate


def residence_time(c, ln, time, delta_t_range=None, range_type='mixed', step=None):
    """Calculate Residence Time. 
    See the following sources for a detailed explanation:
    https://pubs.acs.org/doi/abs/10.1021/ja310577u and https://www.pnas.org/content/117/14/7803

    This metric is currently not calcualted by default. It needs more testing. 
    Its output is also more difficult to control and predict. 

    Parameters
    ----------
    c : contacts
        This is the output of retrieve_contacts_flat, or it can be any flat numpy array.

    ln : int
        The number of the lipid of interest (i.e. the number of residues)

    time : float
        The total trajectory time.

    delta_t_range : list
        The range of the delta_t parameter. If None, it is calcculated.

    range_type : str
        Residence time can be sped up by supplying a default range. You can specify either a geometric,
        linear, or mixed range ('geo', 'mixed', 'linear'). Fastest and default is 'geo'. Only used
        if delta_t_range=None.

    step : int
        Custom step size for delta_t_range. Not implemented.


    """


    from scipy.optimize import curve_fit

    if delta_t_range is None:
        if range_type == 'mixed':
            delta_t_range = list(np.arange(5,100,1)) + list(np.arange(100,1000,5)) + list(np.arange(1000,time,10))
        elif range_type == 'geo':
            delta_t_range = list(np.geomspace(0.01,time,1000))
        elif range_type == 'linear':
            delta_t_range = list(np.arange(5,time,1))

    sigma = {0: 1}
    sigma0 = sum([r for r in c]) / (float(time) * ln)

    for delta_t in delta_t_range:
        denominator = (float(time) - delta_t) * ln * sigma0

        if denominator != 0:
            sigma[delta_t] = sum([r - delta_t for r in c if r >= delta_t]) / denominator
        else:
            sigma[delta_t] = 0

    try:
        popt, pcov = curve_fit(lambda t,a,b,c,d:c*np.exp(-a*t)+d*np.exp(-b*t),
                            list(sigma.keys()),
                            list(sigma.values()),
                            (1, 0.1, 1, 1))
    except RuntimeError:
        return (0, 0)

    return (1 / min([abs(k) for k in popt[:2]]), pcov)



def show_network(filename, width=800, height=600):
    """Visualize lipid-protein contacts using a network graph.

    Parameters
    ----------

    filename : str
        The name of the json file in which network information is stored. This
        is the output of pl.network() call.

    width : int
        The width of the window

    height : int
        The height of the window.

    """

    from IPython.core.display import HTML
    from string import Template, ascii_uppercase, ascii_lowercase, digits
    import json
    import random
    import string

    rand_str = ''.join(random.choices(ascii_uppercase + digits + ascii_lowercase, k=10))

    with open(filename) as f:
        network = json.load(f)

    HTML('<script src="./d3.min.js"></script>')

    js_text_template = Template('''

    var clientHeight = document.getElementById('$network').clientHeight;
    var clientWidth = document.getElementById('$network').clientWidth;

    var root = $python_data ;

    var width = clientWidth,
        height = clientHeight,
        root,
        radius = 6;

    function loadDATA(gpcr) {
        var force = d3.layout.force()
            .size([width, height])
            .on("tick", tick);


        force.charge(function (d) {
            if (d.id <= root.graph.lip_end) {
                return -500
            }
            return -5000
        });


        force.linkDistance(function (d) {
            if ((d.source.id <= root.graph.lip_end) && (d.target.id >= root.graph.res_start + 1)) {
                return 300
            }
            return 180
        });


        var svg = d3.select("#$network").append("svg")
            .attr("width", width)
            .attr("height", height);


        var link = svg.selectAll(".link"),
            node = svg.selectAll(".node");

    function call_root(gpcr) {

        // root = gpcr;

        var a = width / 2
        var b = height / 2


        if (root.graph.lip_end <= 8) {

            var R = 50
            posX = [a, a + R, a - R, a, a, a + (R / 1.41), a + (R / 1.41), a - (R / 1.41), a - (R / 1.41)]
            posY = [b, b, b, b + R, b - R, b + (R / 1.41), b - (R / 1.41), b + (R / 1.41), b - (R / 1.41)]


        } else {

            var R = [50, 70, 90, 110, 30, 130, 140, 160, 180, 200];
            var new_r = 0
            var cp = 1
            var it = 1
            // Hendecagon alignment of lipid nodes
            angle = 0.57
            // angle = 0.785
            var posX = [a];
            var posY = [b];

            while (cp <= root.graph.lip_end) {

                var x = Math.cos(angle*it) * R[new_r];
                var y = Math.sin(angle*it) * R[new_r];
                posX.push(a + x)
                posY.push(b + y)

                if (it < 11) {
                    it++
                } else {
                    var it = 1
                    new_r++
                }
                cp++
            }
    }

        for (var i = 0; i < root.nodes.length; i++) {
            var node = root.nodes[i];

            if (node.id <= root.graph.lip_end) {
                node.x = posX[node.id]
                node.y = posY[node.id]
                node.fixed = true;
            }

            node.id = i;
            node.collapsing = 0;
            node.collapsed = false;
        }

        //Give links ids and initialize variables
        for (var i = 0; i < root.links.length; i++) {
            var link = root.links[i];
            link.source = root.nodes[link.source];
            link.target = root.nodes[link.target];
            link.id = i;
        }

        function hideChildren(d) {
            //check if link is from this node, and if so, collapse
            root.links.forEach(function (l) {
                if ((l.source.id >= root.graph.lip_end + 1) && (l.source.id <= root.graph.res_start)) {
                    l.target.collapsing = 1
                    l.source.collapsed = true
                }
            });

            update();

        }

        for (var i = 0; i < root.nodes.length; i++) {
            hideChildren(node)
        }

        update();

   // });
    };
    call_root(gpcr)

        function update() {

            d3.selectAll('text').remove()

            //Keep only the visible nodes
            var nodes = root.nodes.filter(function (d) {
                return d.collapsing == 0;
            });
            var links = root.links;

            //Keep only the visible links
            links = root.links.filter(function (d) {
                return d.source.collapsing == 0 && d.target.collapsing == 0;
            });


            force
                .nodes(nodes)
                .links(links)
                .friction(0.4)
                .start();

            // Update the links…
            link = link.data(links, function (d) {
                return d.id;
            });

            // Exit any old links.
            link.exit().remove();

            // Enter any new links.
            link.enter().insert("line", ".node")
                .attr("class", "link")
                .attr("x1", function (d) {
                    return d.source.x;
                })
                .attr("y1", function (d) {
                    return d.source.y;
                })
                .attr("x2", function (d) {
                    return d.target.x;
                })
                .attr("y2", function (d) {
                    return d.target.y;
                })
                // .style('stroke-width', 5)
                .style('stroke-width', function (d) {
                    return d.weight * 10;
                })
                .style("stroke-opacity", 0.6)
                .attr("stroke", "#b7b7b7")
                .attr("marker-end", "url(#arrow)");

            // Update the nodes…
            node = node.data(nodes, function (d) {
                return d.id;

            }).style("fill", color);

            // Exit any old nodes.
            node.exit().remove();

            // Enter any new nodes.
            node.enter().append("circle")
                .attr("class", "node")
                .attr("cx", function (d) {
                    return d.x;
                })
                .attr("cy", function (d) {
                    return d.y;
                })
                .attr("r", function (d) {
                    return d.values;
                })
                .style("fill", color)
                .style("cursor", "pointer")
                .on("click", click)
                .on("mouseover", mouseOver(.0))
                .on("mouseout", mouseOut)
                .call(force.drag);

            label = svg.selectAll(null)
                .data(nodes)
                .enter()
                .append("text")
                .text(function (d) {
                    return d.name;
                })
                .style("text-anchor", "middle")
                // .style("fill", "black")
                .style("fill", function(d) {
                    if (d.id <= root.graph.lip_end) {
                        return "#a03271"
                    }
                    return "black"
                })
                .style("font-family", "Arial")
                .style("font-weight", function(d) {
                    if (d.id <= root.graph.lip_end) {
                        return "bold"
                    }
                })
                .style("font-size", function (d) {
                    if (d.id <= root.graph.lip_end) {
                        return 10
                    }
                    return 12
                });

            // fade nodes on hover
            function mouseOver(opacity) {
                return function (d) {

                    var linkedByIndex = {};
                    root.links.forEach(function (d) {
                        linkedByIndex[d.source.id + "," + d.target.id] = 1;
                    });

                    function isConnected(a, b) {
                        return linkedByIndex[a.id + "," + b.id] || linkedByIndex[b.id + "," + a.id] || a.id == b.id;
                    }

                    // check all other nodes to see if they're connected
                    // to this one. if so, keep the opacity at 1, otherwise
                    // fade. Tweak text transparency/opacity
                    label.style("opacity", function (o) {
                        thisOpacity = isConnected(d, o) ? 1 : opacity;
                        return thisOpacity;
                    })

                    node.style("stroke-opacity", function (o) {
                        thisOpacity = isConnected(d, o) ? 0 : opacity;
                        return thisOpacity;
                    });
                    node.style("fill-opacity", function (o) {
                        thisOpacity = isConnected(d, o) ? 1 : opacity;
                        return thisOpacity;
                    });

                    // also style link accordingly
                    link.style("stroke-opacity", function (o) {
                        o.source.colour = "black"
                        return o.source === d || o.target === d ? 0.7 : opacity;
                    });

                    link.style("stroke", function (o) {
                        return o.source === d || o.target === d ? o.source.colour : "#b7b7b7";
                    });
                };
            }

            function mouseOut() {
                node.style("stroke-opacity", 1);
                node.style("fill-opacity", 1);
                link.style("stroke-opacity", 0.5);
                link.style("stroke", "#b7b7b7");
                label.style("opacity", 1);
            }
            // new new
            // if(abc) return;
        }

        function tick() {

            link.attr("x1", function (d) {
                    return d.source.x;
                })
                .attr("y1", function (d) {
                    return d.source.y;
                })
                .attr("x2", function (d) {
                    return d.target.x;
                })
                .attr("y2", function (d) {
                    return d.target.y;
                });

            // Bounding box example: requires radius
            node.attr("cx", function (d) {
                    return d.x = Math.max(radius, Math.min(width - radius, d.x));
                })
                .attr("cy", function (d) {
                    return d.y = Math.max(radius, Math.min(height - radius, d.y));
                });

            node
                .attr("cx", function (d) {
                    return d.x;
                })
                .attr("cy", function (d) {
                    return d.y;
                });

            label
                .attr("x", function (d) {
                    return d.x;
                })
                .attr("y", function (d) {
                    return d.y - 15;
                })

        }

        // Color leaf nodes light blue, and packages grey or black.
        function color(d) {
            if (d.id <= root.graph.lip_end) {
                // return d.collapsed = "#a03271"
                return d.collapsed ? "#d6c8ff" : d.children ? "#c6dbef" : "#a03271";
            }
            return d.collapsed ? "#4b4c7a" : d.children ? "#c6dbef" : "#d6c8ff";
        }

        // Toggle children on click.
        function click(d) {

                //check if link is from this node, and if so, collapse
                root.links.forEach(function (l) {

                    if (d.id <= root.graph.lip_end) {

                        if (d.collapsed) {
                            l.target.collapsing--;
                        } else {
                            l.target.collapsing++;
                        }
                    } else if (l.source.id == d.id) {

                        if (d.collapsed) {

                            l.target.collapsing--;

                        } else {

                            l.target.collapsing++;

                        }
                    }
                });

                d.collapsed = !d.collapsed;

            update();
        }
    }

    function loadGPCR() {
        loadDATA('')
    }
    loadGPCR()
    '''#.format(width, height)
    )

    html_template = Template(
    '<div id=' + rand_str + f' style="width:{width}px; height:{height}px"></div>'
    '''
    <script > $js_text </script>
    ''')

    js_text = js_text_template.substitute({'python_data': json.dumps(network),
                                        'network': rand_str})
    network_plot = html_template.substitute({'js_text': js_text})
    d3_download = """ <script src="http://d3js.org/d3.v3.min.js" charset="utf-8"></script>"""

    network_plot = d3_download + network_plot

    HTML(network_plot)
    return HTML(network_plot)



def shift_range(values, new_min=0, new_max=1):
    """
    Shift values from one range to another.
    """
    old_min = min(values)
    old_max = max(values)

    old_range = (old_max - old_min)
    new_range = (new_max - new_min)
    new_val = []
    for val in values:
        try:
            new_value = (((val - old_min) * new_range) / old_range) + new_min
        except ZeroDivisionError:
            new_value = new_min
        new_val.append(new_value)
    return new_val




from .core.systemtopology import Proteins, Lipids, Protein
from .core.computecontacts import ComputeContacts, contacts_dataframe, retrieve_contacts, retrieve_contacts_flat

from .vis.show_points import show_points
from .vis.show_network import show_network
from .vis.show_contact_projection import show_contact_projection
from .vis.show_distances import show_distances
from .vis.show_radar import show_radar
from .vis.show_metric_distances import show_metric_distances
from .vis.show_density import show_density

from .utils.martinilipids import lipid_species, martini_lipids, martini_headgroups
from .utils.network import network
from .utils.metrics import residence_time

from .utils.misc import save_pdb

from .core.systemtopology import *




import yaml
import numpy as np
import pandas as pd

# bokeh aplication
from bokeh.document import Document

from bokeh.plotting import figure
from bokeh.layouts import layout, column
from bokeh.models import ColumnDataSource, LabelSet
from bokeh.models.widgets import Select
from bokeh.palettes import RdBu

from bokeh.io import show, output_notebook
from bokeh.themes import Theme

from prolintpy.utils.misc import unit_poly_verts, radar_patch, star_curv

def show_radar(df=None, filename=None):
    """Visualize lipid-protein contacts using a Bokeh radar map.

    This is useful if you want to (i) compare the binding profile of two residues and
    (ii) when you want to see how different contact metrics compare against each other.

    Parameters
    ----------

    df : pd.DataFrame
         A contacts df. This will be the output of pl.contacts_dataframe() command.

    filename: string
        A string pointing to the location of a csv file containing data that can be loaded into a pandas DataFrame.

    """
    if df is None and not filename:
        raise Exception("show_radar can take either a pandas DataFrame or a csv file that can be loaded into a pandas DataFrame.")
    elif filename:
        df = pd.read_csv(filename)

    output_notebook()

    def radarApp(doc):

        nonlocal df
        df = df.copy()
        for col in df.columns:
            if col.endswith('Error'):
                del df[col]

        gpcrs = list(df.Protein.unique())
        lipids = list(df.Lipids.unique())
        radius = list(df.Radius.unique())
        radius = [str(r) for r in radius]

        # Normalize [0-1] all contact parameters with respect to each other.
        for cmn in list(df.columns)[:-5]:
            for lipid in lipids:
                for radius_value in radius:
                    radius_value = float(radius_value)

                    dfmin = df[
                        (df.Lipids == lipid) &
                        (df.Radius == radius_value)
                        ][cmn].min()

                    dfmax = df[
                        (df.Lipids == lipid) &
                        (df.Radius == radius_value)
                        ][cmn].max()

                    lip_index = list(df[
                        (df.Lipids == lipid) &
                        (df.Radius == radius_value)
                        ].index)

                    df.loc[lip_index, cmn] = df[
                        (df.Lipids == lipid) &
                        (df.Radius == radius_value)
                        ][cmn].apply(lambda x: (x-dfmin) / (dfmax - dfmin))

        df = df.fillna(0)


        # 1 unique set of residues - works also if there's only 1 set available.
        resname = df[(df.Protein == gpcrs[0]) & (df.Lipids == lipids[0]) & (df.Radius == float(radius[0]))].ResName.to_list()
        resid   = df[(df.Protein == gpcrs[0]) & (df.Lipids == lipids[0]) & (df.Radius == float(radius[0]))].ResID.to_list()

        residues = []
        for rn, ri in zip(resname, resid):
            residues.append("{}-{}".format(rn, ri))


        import random
        # Create Input controls
        gpcr1 = Select(title="Protein", value=gpcrs[0], options=gpcrs, width=100)
        gpcr2 = Select(title="Protein", value=gpcrs[0], options=gpcrs, width=100)
        lipid1 = Select(title="Lipid", value=lipids[0], options=lipids, width=100)
        lipid2 = Select(title="Lipid", value=lipids[0], options=lipids, width=100)
        residue1 = Select(title="Residue", value=random.choice(residues), options=residues, width=120)
        residue2 = Select(title="Residue", value=random.choice(residues), options=residues, width=120)
        radius1 = Select(title="Radius", value=radius[0], options=radius, width=100)
        radius2 = Select(title="Radius", value=radius[0], options=radius, width=100)


        # num_vars = 6
        # cols_to_display = [x for x in list(df.columns) if not x.endswith('Error')]
        # num_vars = len(cols_to_display)

        num_vars = len(list(df.columns)[:-5])
        centre = 0.5

        theta = np.linspace(0, 2*np.pi, num_vars, endpoint=False)
        theta += np.pi/2

        verts = unit_poly_verts(theta, centre)
        xv = [v[0] for v in verts]
        yv = [v[1] for v in verts]

        f1 = df[
                (df.Protein == str(gpcr1.value)) &
                (df.Lipids == str(lipid1.value)) &
                (df.ResID == int(str(residue1.value).split("-")[1])) &
                (df.Radius == float(radius1.value))
                ].loc[:, list(df.columns)[:num_vars]].to_numpy().reshape(num_vars,)

        f2 = df[
                (df.Protein == str(gpcr2.value)) &
                (df.Lipids == str(lipid2.value)) &
                (df.ResID == int(str(residue2.value).split("-")[1])) &
                (df.Radius == float(radius2.value))
                ].loc[:, list(df.columns)[:num_vars]].to_numpy().reshape(num_vars,)

        flist = [f1, f2]

        p = figure(plot_height=400, plot_width=800, x_range=(-0.3, 1.3), y_range=(-0.3, 1.3), tools="")
        cmap = RdBu[num_vars]
        colors = cmap[:1] + cmap[-1:]

        patch1 = ColumnDataSource(data=dict(xi=[], yi=[]))
        circle1 = ColumnDataSource(data=dict(xt=[], yt=[]))
        xt, yt = radar_patch(f1, theta, centre)
        xi, yi = star_curv(xt + xt[:1], yt + yt[:1])
        patch = p.patch(x="xi", y="yi", fill_alpha=0.15, fill_color=colors[0], line_color=colors[0], source=patch1)
        circle = p.circle(x='xt', y='yt', size=10, alpha=0.5, line_color=colors[0], fill_color=colors[0], source=circle1)

        patch2 = ColumnDataSource(data=dict(xi=[], yi=[]))
        circle2 = ColumnDataSource(data=dict(xt=[], yt=[]))
        xt, yt = radar_patch(f2, theta, centre)
        xi, yi = star_curv(xt + xt[:1], yt + yt[:1])
        patch = p.patch(x="xi", y="yi", fill_alpha=0.15, fill_color=colors[1], line_color=colors[1],  source=patch2)
        circle = p.circle(x='xt', y='yt', size=10, alpha=0.5, line_color=colors[1], fill_color=colors[1], source=circle2)

        # Draw grid so the plot looks like a polar graph.
        text_label = list(df.columns)[:num_vars] + ['']
        p.circle(x=0.5, y=0.5, size=20, fill_alpha=0, line_color='black', line_dash='dashed', line_alpha=0.2)
        for size in [x*100 for x in range(1, 10)]:
            p.circle(x=0.5, y=0.5, size=size, fill_alpha=0, line_color='black', line_dash='dashed', line_alpha=0.2)

        line = np.linspace(-2, 2, 100)
        p.line(line, (line * 0.58)+0.21, line_color='black', line_dash='dashed', line_alpha=0.2)
        p.line(line, (line * (-0.58))+0.79, line_color='black', line_dash='dashed', line_alpha=0.2)
        p.line([0.5]*100, line, line_color='black', line_dash='dashed', line_alpha=0.2)

        # Hide axes.
        p.xgrid.grid_line_color = None
        p.ygrid.grid_line_color = None
        p.xaxis.major_tick_line_color = None
        p.xaxis.minor_tick_line_color = None
        p.yaxis.major_tick_line_color = None
        p.yaxis.minor_tick_line_color = None
        p.xaxis.major_label_text_font_size = '0pt'
        p.yaxis.major_label_text_font_size = '0pt'

        p.axis.visible = None
        p.toolbar.logo = None
        p.toolbar_location = None

        # Draw the hexagon with labels attached.
        p.line(x=xv+[centre], y=yv+[1], color='black')

        xv[1] = xv[1] - 0.13
        xv[2] = xv[2] - 0.1
        xv[-2] = xv[-2] + 0.16
        xv[-1] = xv[-1] + 0.1
        # yv[3] = yv[3] - 0.06

        label_source = ColumnDataSource({'x':xv + [centre ],'y':yv + [1],'text':text_label})
        labels = LabelSet(x="x",y="y",text="text",source=label_source, level='glyph', text_align="center")
        p.add_layout(labels)

        def select_dfGPCR():
            gpcr1_val, gpcr2_val  = gpcr1.value, gpcr2.value
            lipid1_val, lipid2_val = lipid1.value, lipid2.value
            res1_val, res2_val = str(residue1.value).split("-")[1], str(residue2.value).split("-")[1]

            f1 = df[
                (df.Protein == str(gpcr1_val)) &
                (df.Lipids == str(lipid1_val)) &
                (df.ResID == int(res1_val)) &
                (df.Radius == float(radius1.value))
                ].loc[:, list(df.columns)[:num_vars]].to_numpy().reshape(num_vars,)
            f2 = df[
                (df.Protein == str(gpcr2_val)) &
                (df.Lipids == str(lipid2_val)) &
                (df.ResID == int(res2_val)) &
                (df.Radius == float(radius2.value))
                ].loc[:, list(df.columns)[:num_vars]].to_numpy().reshape(num_vars,)

            return (f1, f2)

        def update1():
            gpcr_val = gpcr1.value
            lipid1_val = lipid1.value
            radius1_val = radius1.value

            resname = df[(df.Protein == str(gpcr_val)) & (df.Lipids == str(lipid1_val)) & (df.Radius == float(radius1_val))].ResName.to_list()
            resid   = df[(df.Protein == str(gpcr_val)) & (df.Lipids == str(lipid1_val)) & (df.Radius == float(radius1_val))].ResID.to_list()

            residues = []
            for rn, ri in zip(resname, resid):
                residues.append("{}-{}".format(rn, ri))

            residue1.options = residues
            if residue1.value not in residues:
                residue1.value = residues[0]

            f1 = select_dfGPCR()[0]
            xt, yt = radar_patch(f1, theta, centre)
            xi, yi = star_curv(xt + xt[:1], yt + yt[:1])

            circle1.data=dict(xt=xt, yt=yt)
            patch1.data=dict(xi=xi, yi=yi)

        def update2():
            gpcr_val = gpcr1.value
            lipid2_val = lipid2.value
            radius2_val = radius2.value

            resname = df[(df.Protein == str(gpcr_val)) & (df.Lipids == str(lipid2_val)) & (df.Radius == float(radius2_val))].ResName.to_list()
            resid   = df[(df.Protein == str(gpcr_val)) & (df.Lipids == str(lipid2_val)) & (df.Radius == float(radius2_val))].ResID.to_list()

            residues = []
            for rn, ri in zip(resname, resid):
                residues.append("{}-{}".format(rn, ri))

            # options = [str(x) for x in options]
            residue2.options = residues
            if residue2.value not in residues:
                residue2.value = residues[0]

            f2 = select_dfGPCR()[1]
            xt, yt = radar_patch(f2, theta, centre)
            xi, yi = star_curv(xt + xt[:1], yt + yt[:1])

            circle2.data=dict(xt=xt, yt=yt)
            patch2.data=dict(xi=xi, yi=yi)


        controls1 = [gpcr1, lipid1, residue1, radius1]
        controls2 = [gpcr2, lipid2, residue2, radius2]

        for control1 in controls1:
            control1.on_change('value', lambda attr, old, new: update1())
        for control2 in controls2:
            control2.on_change('value', lambda attr, old, new: update2())

        inputs1 = column([gpcr1, lipid1, residue1, radius1])
        inputs2 = column([gpcr2, lipid2, residue2, radius2])

        l = layout([[inputs1, p, inputs2]])

        update1()
        update2()

        doc.add_root(l)

        doc.title = "Radar App"
        doc.theme = Theme(json=yaml.load("""
            attrs:
                Figure:
                    toolbar_location: above
                    height: 500
                    width: 800
                Grid:
                    grid_line_dash: [6, 4]
                    grid_line_color: black
        """, Loader=yaml.FullLoader))

    return show(radarApp)



import re
import numpy as np
import pandas as pd

class SystemTopology(object):
    """Base class for Lipids and Proteins classes to inherit from.

    It reads an MDTraj topology and extracts useful information. Its main
    purpose is intended to be in distinguishing the resolution of the input
    system. Other ProLint classes and functions should ideally be
    resolution-agnostic, though currently that is not entirely the case.

    ProLint, in its current version, only supports the Martini model.


    Attributes
    ----------
    topology : MDTraj.Topology

    resolution : Either "martini" or "atomistic".

    """

    def __init__(self, topology, resolution="martini"):
        """Create a new ProLint topology."""

        if resolution == "martini":
            lipid_indices = topology.select("all and not (name BB or (name =~ 'SC[1-9]'))")
            proteins_indices = topology.select("name BB or (name BB or (name =~ 'SC[1-9]'))")
        elif resolution == "atomistic":
            lipid_indices = topology.select("all and not protein")
            proteins_indices = topology.select("protein")
        else:
            raise NotImplementedError('Provided resolution is not supported. Only "martini" and "atomistic" are supported.')

        self.resolution = resolution
        self.topology = topology
        self.dataframe = topology.to_dataframe()[0]
        self.l_indices = lipid_indices
        self.p_indices = proteins_indices
        self.pdf = self.dataframe.iloc[proteins_indices]
        self.ldf = self.dataframe.iloc[lipid_indices]


    def __str__(self):
        return "Base Class for the topology of the MDTraj system."

    def __repr__(self):
        return "Base Class for the topology of the MDTraj system."


class Lipids(SystemTopology):
    """
    Lipids stores information about the lipid composition of the system.
    Plasma membrane lipids (as described here: https://pubs.acs.org/doi/abs/10.1021/acscentsci.8b00143)
    work best, but all lipids in the Martin lipidome should work (see here:
    http://www.cgmartini.nl/index.php/force-field-parameters/lipids)

    g_surf only works with Plasma lipids.

    Attributes
    ----------
    topology : MDTraj.Topology
        MDTraj.Topology instance

    lipid_names : list
        Only select the following lipids. Useful when you want to analyze contacts
        with only one or a few lipids instead of all system lipids.

    atom_names : list
        Only select the following atoms/beads. Useful when you want to further
        restrict analysis to specific bead/atom types instead of whole lipid residue.

    """

    def __init__(self, topology, resolution="martini", lipid_names=None, atom_names=None):
        """Create a new ProLint Lipids class."""

        super().__init__(topology, resolution=resolution)

        lfunc = lambda x, y: True if y is None else x in y
        user_indices = [a.index for a in topology.atoms
                        if lfunc(a.residue.name, lipid_names)
                        and lfunc(a.name, atom_names)
                        ]

        if len(user_indices) == 0:
            print ("No lipids have been selected!.")

        self.l_indices = np.intersect1d(self.ldf.index.to_numpy(), user_indices)
        self.dataframe = self.dataframe.iloc[self.l_indices]


    def lipid_names(self):
        """Get the names of all lipids that will be analyzed.

        Returns
        -------
        names : array of lipid names
        """
        return self.dataframe.resName.unique()

    def lipid_count(self):
        """Get the name and count of each lipid that will be analyzed.

        Returns
        -------
        lipic_count : dictionary
            key:value corresponds to lipid_name:count.
        """
        lc = {}
        lipids = self.lipid_names()
        for lipid in lipids:
            lc[lipid] = self.dataframe[self.dataframe.resName == lipid].resSeq.unique().size
        return lc


    def __str__(self):
        return "<prolintpy.Lipids containing {} atom(s)>".format(self.l_indices.size)

    def __repr__(self):
        return "<prolintpy.Lipids containing {} atom(s)>".format(self.l_indices.size)


class Protein(object):
    """
    Store various information for each protein in the system.

    Attributes
    ----------
    name : str
        The name of the protein.

    """


    def __init__(self, name):
        """Create a new ProLint Protein class."""
        self.name = name
        self.counter = 1
        self.dataframe = []
        self.beads = 0
        self.n_residues = 0
        self.residues = None
        self.resnames = None
        self.chains = 1
        self.resolution = None

    def get_indices(self, residues=None, df=None, copy=None, resolution='martini', suppress=False):
        """Return the indices for each protein residue. If there are more than one
        protein replicates/copies in the system the method will choose the residue indices of
        the first occuring protein. Alternatively you can provide the dataframe of the protein
        copy yourelf using the 'dataframe' attribute of this class.

        Parameters
        ----------
        residues : list
            Explicitly specify the residues. List should be 1-based, meaning the residue numbers should
            correspond to residues in the input topology.
        df : DataFrame
            The protein dataframe. It can be a custom definition, but usually you would want
            one of the dataframes contined in the self.dataframe attribute.
        copy : int
            Specify the protein copy/replicate.
        resolution : str
            The resolution of the system. Either "martini" or "atomistic".
        suprress : bool, default=True
            Suppress warning messages.

        Returns
        -------
        indices : list
            list of numpy arrays.
        """
        if df is None:
            if not suppress:
                if len(self.dataframe) > 1 and copy is None:
                    print (re.sub(' +', ' ', """Object has more than one dataframe. Using the first one.
                            Alternatively, you can specify the dataframe of the protein yourself using the
                            dataframe attribute of this class."""))
                else:
                    print ("Using the available dataframe")
            df=self.dataframe[0]

        if copy is not None:
            df=self.dataframe[copy]

        if residues is None:
            if resolution == "martini":
                residues = df[df.name == "BB"].resSeq.to_list()
            elif resolution == "atomistic":
                residues = df[df.name == "CA"].resSeq.to_list()
            else:
                raise ValueError("Only martini and atomistic are currently supported.")

            residues = residues[:int(len(residues) / self.chains)]
        else:
            assert all(x in self.residues for x in residues), "Supplied residues should correspond to input topology residues."

        # if self.chains > 1:
        #     residues = residues[:int(len(residues) / self.chains)]

        indices = [df[df.resSeq == x].index.to_numpy() for x in residues]

        return indices

    def __str__(self):
        return "<prolintpy.Protein containing {} replicate(s) of {} and {} beads each>".format(self.counter, self.name, self.n_residues)

    def __repr__(self):
        return "<prolintpy.Protein containing {} replicate(s) of {} and {} beads each>".format(self.counter, self.name, self.n_residues)


class Proteins(SystemTopology):
    """
    Proteins stores information about the protein composition of the system.

    Gromacs coordinate files do not contain information on protein names and their
    count. This class tries to calculate and extract this information from the topology.

    The class has a 'system_proteins' method that can be used to retrieve information
    about each protein specifically.

    Attributes
    ----------
    topology : MDTraj.Topology
        An MDTraj.Topology instance.

    """

    def __init__(self, topology, resolution="martini"):
        """Create a new ProLint Proteins class."""

        super().__init__(topology, resolution=resolution)

        # Get start and end indices of proteins in the system.
        # The assumption here is that proteins are ordered and the start residue of the next
        # protein is always smaller than the last residue of the previous protein.
        resseq = self.pdf.resSeq.to_list()
        p0 = resseq[0]
        # system_proteins contains the start and end indices of all proteins.
        fi_li = []
        fi = 0
        for li, p in enumerate(resseq):
            if p < p0:
                fi_li.append((fi, li-1))
                fi = li
            p0 = p
        fi_li.append((fi, li))

        self.fi_li = fi_li

    def system_proteins(self, merge=True):
        """Stores information for each protein in the system using the Protein class.

        Returns
        -------
        proteins : list
            list of Protein classes. One for each different protein in the system.

        """
        if self.resolution == "martini":
            ref_atom = 'BB'
        elif self.resolution == "atomistic":
            ref_atom = 'CA'

        c = 0
        proteins = []
        # Two proteins are the same if residue number and all beads are equal between them.
        for values in self.fi_li:

            # first and last index
            fi = values[0]
            li = values[1]

            # Get the data for the current protein.
            current_protein_df = self.pdf[(self.pdf.index >= fi) & (self.pdf.index <= li)]
            curr_len = len(current_protein_df[current_protein_df.name == ref_atom])

            # curr_len = len(current_protein_df)
            curr_names = current_protein_df.name.to_list()

            new_protein = True
            if merge:
                for pc in proteins:
                    if curr_names == pc.beads:
                        pc.counter += 1
                        pc.dataframe.append(current_protein_df.copy())
                        new_protein = False

            if new_protein:
                protein = Protein(f'Protein{c}')
                protein.dataframe.append(current_protein_df.copy())
                protein.beads = curr_names
                protein.n_residues = curr_len
                protein.residues = current_protein_df.resSeq.unique()

                resnames = current_protein_df[current_protein_df.name == ref_atom].resName.to_numpy()
                protein.resnames = resnames

                protein.first_residue = current_protein_df.resSeq.to_list()[0]
                protein.last_residue = current_protein_df.resSeq.to_list()[-1]

                protein.resolution = self.resolution

                proteins.append(protein)

            c += 1

        return proteins

    def merge_chains(self, proteins):
        """If the input structure file contained multiple and similar chains, ProLint
        will count them as separate proteins by default. This function allows you to
        change that. You call it on the list of proteins that you want to change and it will
        merge those proteins into multiple chains.

        Parameters
        ----------
        proteins : ProLint.Proteins
            Proteins object.

        """
        merged = []
        for protein in proteins:
            protein.beads = protein.beads * protein.counter
            protein.dataframe = [pd.concat(protein.dataframe)]
            # protein.n_residues = protein.n_residues #* protein.counter => n_residues remains the same
            protein.chains = protein.counter
            protein.counter = 1
            merged.append(protein)
            print ("Merged copies as chains in {}".format(protein.name))

        return merged




    def __str__(self):
        return "class storing information on system proteins"

    def __repr__(self):
        return "class storing information on system proteins"



import os
import json
import networkx as nx
from networkx.readwrite import json_graph

from prolintpy.utils.shift_range import shift_range

def network(l, df, metric, grouped, radius=None, return_json=False):
    """Calculate network information.

    ** The code has to be updated and improved for stability and readability **

    Parameters
    ----------

    l : ProLint.Lipids
        System lipids class.

    df : pd.DataFrame
        pandas dataframe of contacts

    metric : str
        The metric to for network analysis.

    grouped : bool
        Are lipid grouped into headgroup categories.

    radius : float
        Cutoff radius used in measuring contacts. Useful when there's multiple
        cutoff radii.

    return_json : bool
        Return the results instead of saving them to a file.

    """

    RESIDUES = ["ALA", "ARG", "ASN", "ASP", "CYS", "GLU", "GLN",
                "GLY", "HIS", "ILE", "LEU", "LYS", "MET", "PHE",
                "PRO", "SER", "THR", "TRP", "TYR", "VAL"]

    if grouped:
        from prolintpy.utils.martinilipids import martini_lipids
        PLASMA_LIPIDS = martini_lipids(l.lipid_names())
    else:
        PLASMA_LIPIDS = {}
        for lip in l.lipid_names():
            PLASMA_LIPIDS[lip] = [lip]

    if radius is None:
        if df.Radius.unique().size != 1:
            raise ValueError("Ambiguous cutoff. Specify radius parameter.")
        r = df.Radius.unique()[0]
    else:
        r = radius

    ALL_LIPIDS = []
    for lip in l.lipid_names():
        lipid_df = l.ldf
        lipid_top_bead = lipid_df[(lipid_df.resName == lip)].name.unique()[0]
        c = lipid_df[(lipid_df.resName == lip) & (lipid_df.name == lipid_top_bead)].count().serial
        ALL_LIPIDS.append((lip, c))


    proteins = df.Protein.unique()
    LIPIDS = df.Lipids.unique()

    def get_data_network(r, proteins, df, metric):
        """
        Getting the data.
        """
        data = {}
        for gpcr in proteins:
            gpcr_df = df[(df.Radius == float(r)) & (df.Protein == gpcr)]
            lipids = gpcr_df.Lipids.unique()
            lipid_number = len(lipids)

            lipid_dict = {}

            for lipid in lipids:
                lipid_df = gpcr_df[gpcr_df.Lipids == lipid]
                residue_dict = {}

                for residue in RESIDUES:
                    residue_df = lipid_df[lipid_df.ResName == residue]
                    value = residue_df[residue_df[metric] > 0][metric].mean()
                    if value > 0:
                        residue_dict[residue] = value

                lipid_dict[lipid] = residue_dict

            data[gpcr] = lipid_dict


        data_dict = {}

        for gpcr in proteins:
            gpcr_df = df[(df.Radius == float(r)) & (df.Protein == gpcr)]
            lipids = gpcr_df.Lipids.unique()
            lipid_number = len(lipids)
            lipid_dict = {}

            for lipid in lipids:
                lipid_df = gpcr_df[gpcr_df.Lipids == lipid]

                res_names = lipid_df.ResName.to_list()
                res_numbers = lipid_df.ResID.to_list()
                parameters = lipid_df[metric].to_list()

                residue_dict = {}

                for res_name, res_number, parameter in zip(res_names, res_numbers, parameters):
                    name_number = res_name + "-" + str(res_number)
                    residue_dict[str(name_number)] = parameter

                lipid_dict[lipid] = residue_dict

            data_dict[gpcr] = lipid_dict

        return data, data_dict, lipid_number


    def protein_lipid_network(prot, data, data_dict, lipid_number, r, df, ALL_LIPIDS, PLASMA_LIPIDS):
        """
        Main function to calculate the protein network.
        """

        protein_df = df[(df.Radius == float(r)) & (df.Protein == prot)]
        lipids = protein_df.Lipids.unique()

        residue_id_dict = data[prot]
        protein_dict = data_dict[prot]

        residues = protein_df[protein_df.Lipids == lipids[0]].ResName
        residues_number = len(residues)
        residue_ratios = {}
        for amino_acid in RESIDUES:
            residue_number = len(residues[residues == amino_acid])
            residue_ratios[amino_acid] = residue_number/residues_number

        # this is the same as lipid_names
        system_lipids = [x[0] for x in dict(ALL_LIPIDS).items()]

        # ALL_LIPIDS may contain the same lipid twice (for each lipid).
        # When calculating the content ratio of lipids we have to consider this.
        total_lipids = 0
        lip_size = {}
        for lipid_name in system_lipids:
            leaflet_lipids = [x for x in ALL_LIPIDS if x[0] == lipid_name]
            lipid_sum = 0
            leaflet_lipid = None
            for leaflet_lipid in leaflet_lipids:
                lipid_sum += leaflet_lipid[1]
            total_lipids += lipid_sum
            if leaflet_lipid is not None:
                lip_size[leaflet_lipid[0]] = lipid_sum


        lipid_sum = 0
        lip_size_dict = {}
        for lipid_group, leaflet_lipid in PLASMA_LIPIDS.items():
            for lip in leaflet_lipid:
                if lip_size.get(lip):
                    lipid_sum += lip_size[lip]
            try:
                lip_size_dict[lipid_group] = lipid_sum/total_lipids
            except ZeroDivisionError:
                lip_size_dict[lipid_group] = 0
            lipid_sum = 0

        # We just need one lipid to slice the data frame to get the
        # residue name and number.
        residue_number_name = []
        for index in range(len(residues)):
            name = protein_df[protein_df.Lipids == lipids[0]].ResName.to_list()[index]
            number = protein_df[protein_df.Lipids == lipids[0]].ResID.to_list()[index]
            residue_number_name.append((name, number))

        # Node names and IDs
        node_number = list(lipids) + list(RESIDUES) + residue_number_name
        core_nodes = len(lipids) + len(RESIDUES)
        node_dict = {}
        count1, count2 = 0, 0
        for node in node_number:

            if count1 >= core_nodes:
                k = residue_number_name[count2]
                res_id = k[0] + "-" + str(k[1])
                node_dict[res_id] = count1
                count1 += 1
                count2 += 1
                continue

            node_dict[node] = count1
            count1 += 1


        exclusions = []
        for residue in RESIDUES:
            for name_number in residue_number_name:
                if residue == name_number[0]:

                    counter = 0
                    for lipid in lipids:
                        res_id = name_number[0] + "-" + str(name_number[1])
                        if protein_dict[lipid][res_id] > 0.05:
                            break
                        else:
                            counter += 1

                    if counter == lipid_number:
                        exclusions.append(res_id)


        node_keys, node_vals = [], []
        node_id = 0
        for node_key in node_dict.items():
            if node_key[0] not in exclusions:
                node_keys.append(node_key[0])
                node_vals.append(node_id)
                node_id += 1

        node_dict = dict(zip(node_keys, node_vals))

        # LINKING NODES
        edge_data = []
        for lipid in lipids:
            for residue in RESIDUES:
                try:
                    edge_data.append((node_dict[lipid],
                                    node_dict[residue],
                                    residue_id_dict[lipid][residue]))
                except KeyError:
                    edge_data.append((node_dict[lipid], node_dict[residue], 0))


        for residue in RESIDUES:
            for name_number in residue_number_name:
                if residue == name_number[0]:
                    for lipid in lipids:
                        res_id = name_number[0] + "-" + str(name_number[1])
                        if protein_dict[lipid][res_id] > 0.05:
                            edge_data.append((node_dict[residue], node_dict[res_id], 0.1))
                            break

        for lipid in lipids:
            for name_number in residue_number_name:
                res_id = name_number[0] + "-" + str(name_number[1])
                if protein_dict[lipid][res_id] > 0.05:
                    edge_data.append((node_dict[lipid], node_dict[res_id], protein_dict[lipid][res_id]))



        edges_tmp = []
        for edge in edge_data:

            if (edge[0] >= lipid_number) & (edge[0] < lipid_number) & (edge[1] >= lipid_number):

                for key, value in node_dict.items():
                    if value == edge[1]:
                        node_id = key

                if node_id not in exclusions:
                    edges_tmp.append(edge)

            else:
                edges_tmp.append(edge)


        # print (edges_tmp)
        weights = [x[2] for x in edges_tmp]
        weights = shift_range(weights, new_min=0.05, new_max=2)
        edges_tmp = [(x[0], x[1], weights[i]) for i, x in enumerate(edges_tmp)]
        edges = []
        for index in enumerate(edges_tmp):
            edges.append((edges_tmp[index[0]][0],
                        edges_tmp[index[0]][1],
                        {"weight": edges_tmp[index[0]][2]}))


        # Set new ranges.
        r_r = [residue_ratios[x] for x in RESIDUES]
        # node_values_1 = shift_range(min(r_r), max(r_r), r_r)
        node_values_1 = shift_range(r_r, new_min=5, new_max=15)

        l_r = [lip_size_dict[x] for x in lipids]
        # node_values_2 = shift_range(min(l_r), max(l_r), l_r)
        node_values_2 = shift_range(l_r, new_min=5, new_max=15)

        node_values = node_values_2 + node_values_1
        extend = len(node_vals) - len(node_values)
        default_size = 15 - 5
        node_values.extend([default_size] * extend)

        return (node_vals, node_values, node_keys, edges)


    data, data_dict, lipid_number = get_data_network(r, proteins, df, metric)
    collect_results = []
    for gpcr in proteins:

        results = protein_lipid_network(gpcr, data, data_dict, lipid_number, r, df, PLASMA_LIPIDS, PLASMA_LIPIDS)
        node_vals, node_values, node_keys, edges = results

        graph_nx = nx.Graph(lip_end=lipid_number-1, res_start=20+lipid_number-1)
        for i in range(len(node_vals)):
            graph_nx.add_node(i, values=node_values[i], color="red", name=node_keys[i])

        graph_nx.add_edges_from(edges)
        graph_json = json_graph.node_link_data(graph_nx)

        collect_results.append((graph_json, gpcr))

    if return_json:
        return collect_results

    for result in collect_results:
        graph_json, gpcr = result
        file_name = os.path.join(gpcr + '_' + metric + "_" + str(r) + '_network.json')
        json.dump(graph_json, open(file_name, 'w'))

    return None



import numpy as np
import mdtraj as md
import matplotlib as mpl
from matplotlib.pyplot import cm

import nglview as nv

from prolintpy.utils.shift_range import shift_range

def show_contact_projection(t, bf, protein=None, residue_list=None, ngl_repr='surface', cmap='Reds'):
    """Visualize lipid-protein contacts by mapping them onto the structure of the protein.

    Parameters
    ----------

    t : MDTraj.Trajectory

    bf : list
        List of contacts.

    protein : ProLint.Protein
    
    residue_list: list
        List of residues. Required when working with a subset of residues. 

    ngl_repr: str
        One representation that will be used by nglview to display the protein. The following are supported
        options: 'point', 'line', 'rope', 'tube', 'trace', 'label', 'cartoon', 'licorice', 'ribbon',
        'backbone', 'spacefill', 'ball+stick'

    cmap : str
        One of matplotlib cmaps to be used to color contacts.

    """

    if protein:
        first_frame = t[0]
        df = protein.dataframe[0]

        if protein.resolution == "martini":
            indices = df[(df.name == "BB")].index.to_numpy()
        elif protein.resolution == "atomistic":
            indices = df[(df.name == "CA")].index.to_numpy()

    else:
        df = t.topology.to_dataframe()[0]
        indices = df.index.to_numpy()

    t_slice = t[0].atom_slice(indices)
    resseq = df.resSeq.to_list()
    
    
    atomic_bfactors = []
    if residue_list is not None:
        init, bfactors = 1, []
        for idx, residue in enumerate(residue_list):
            while init < residue:
                bfactors.append(0)
                init += 1
            bfactors.append(bf[idx])
            init += 1
        dif = indices.size - len(bfactors)
        for remaining_residues in range(dif):
            bfactors.append(0)
            
        atomic_bfactors = []
        for atom in resseq:
                atomic_bfactors.append(bfactors[atom-1])
    else:
        if len(df.resSeq.unique()) != len(bf):
            raise TypeError ('When projecting only a subset of residues provide a list of tuples: [(residue_id, value), ...]')
        for atom in resseq:
            atomic_bfactors.append(bf[atom-1])

    bf_cmap = cm.get_cmap(cmap)

    colors = [mpl.colors.to_hex(x) for x in bf_cmap(shift_range(atomic_bfactors))]
    cs = [[y, str(resseq[x])] for x, y in enumerate(colors)]

    scheme = nv.color._ColorScheme(cs, 'bf')

    reprs = [
        'point', 'line', 'rope', 'tube', 'trace', 'label',
        'cartoon', 'licorice', 'ribbon',
        'backbone', 'spacefill', 'ball+stick'
        ]


    params = {
        'surface' : [{'type': 'surface',
                      'params': {'surfaceType': "av",
                                 'probeRadius': 2.1,
                                 'color':scheme}}],
     }
    for rep in reprs:
        params[rep] = [{'type': rep, 'params': {'color':scheme}}]

    view = nv.NGLWidget()

    view.add_component(t_slice)
    view.set_representations(params[ngl_repr])

    return view




import mdtraj as md
import numpy as np
from scipy.interpolate import interp1d
from scipy.spatial import ConvexHull

def save_pdb(t, protein, filename, bfactors=None):
    """Save a PDB file. It supports adding contact information
    in a beta-factor column.

    Parameters
    ----------

    t : MDTraj.Trajectory

    protein : ProLint.Protein

    filename : str

    bfactors : array

    """

    df = protein.dataframe[0]
    if protein.resolution == 'atomistic':
        indices = df[(df.name == "CA")].index.to_numpy()
    elif protein.resolution == 'martini':
        indices = df[(df.name == "BB")].index.to_numpy()

    t = t[0].atom_slice(indices)

    t.save_pdb(filename, bfactors=bfactors)


def unit_poly_verts(theta, centre ):
    """Return vertices of polygon for subplot axes.
    This polygon is circumscribed by a unit circle centered at (0.5, 0.5)
    """
    x0, y0, r = [centre ] * 3
    verts = [(r*np.cos(t) + x0, r*np.sin(t) + y0) for t in theta]
    return verts

def radar_patch(r, theta, centre ):
    """ Returns the x and y coordinates corresponding to the magnitudes of
    each variable displayed in the radar plot
    """
    # offset from centre of circle
    offset = 0
    yt = (r*centre + offset) * np.sin(theta) + centre
    xt = (r*centre + offset) * np.cos(theta) + centre
    return list(xt), list(yt)



def star_curv(old_x, old_y):
    """ Interpolates every point by a star-shaped curv. It does so by adding
    "fake" data points in-between every two data points, and pushes these "fake"
    points towards the center of the graph (roughly 1/4 of the way).
    """

    def midpoint(p1, p2, sf=1):
        xm = ((p1[0]+p2[0])/2) * sf
        ym = ((p1[1]+p2[1])/2) * sf
        return (xm, ym)

    try:
        points = np.array([old_x, old_y]).reshape(7, 2)
        hull = ConvexHull(points)
        x_mid = np.mean(hull.points[hull.vertices,0])
        y_mid = np.mean(hull.points[hull.vertices,1])
    except:
        x_mid = 0.5
        y_mid = 0.5

    c=1
    x, y = [], []
    for i, j in zip(old_x, old_y):
        x.append(i)
        y.append(j)
        try:
            xm_i, ym_i = midpoint((i, j),
                                midpoint((i, j), (x_mid, y_mid)))

            xm_j, ym_j = midpoint((old_x[c], old_y[c]),
                                midpoint((old_x[c], old_y[c]), (x_mid, y_mid)))

            xm, ym = midpoint((xm_i, ym_i), (xm_j, ym_j))
            x.append(xm)
            y.append(ym)
            c += 1
        except IndexError:
            break


    orig_len = len(x)
    x = x[-3:-1] + x + x[1:3]
    y = y[-3:-1] + y + y[1:3]


    t = np.arange(len(x))
    ti = np.linspace(2, orig_len + 1, 10 * orig_len)

    kind='quadratic'
    xi = interp1d(t, x, kind=kind)(ti)
    yi = interp1d(t, y, kind=kind)(ti)

    return xi, yi








import numpy as np
import pandas as pd
import mdtraj as md
from collections import defaultdict, Counter
from prolintpy.utils.martinilipids import martini_lipids, lipid_species, martini_headgroups
from prolintpy.utils.metrics import FUNCS, calculate_metric, residence_time
# from prolintpy.core.systemtopology import Lipids

class ComputeContacts(object):
    """Stores information on the contact analysis between system proteins and lipids.

    Instantiating the class only creates the object and populates a few attributes. To calculate
    and analyze contacts you have to call the compute_neighbors method which is the most time
    consuming part and depending on the input parameters can take considerable time to finish.

    Attributes
    ----------
    t : MDTraj.Trajectory

    proteins : ProLint.Proteins.system_proteins()

    lipids : ProLint.Lipids

    """

    def __init__(self, t, proteins, lipids, timestep=None):
        """Create new ProLint Contacts class"""

        """

        timestep is used to convert contacts into trajectory time units. The simplest case,
        when stride=1, is to multiply contacts with dt. When a stride > 1 is used, however, the
        exact time cannot be retrieved. ProLint can estimate that using either:
        n_frames * stride - stride, or using (n_frames * stride) - (stride - 1).

        """

        # TODO:
        # assert that proteins is a list

        self.frames = (t.n_frames - 1)
        self.timestep = t.timestep
        self.t_unit = "unit"
        self.dt = None

        self.proteins = proteins
        self.lipids = lipids

    def compute_neighbors(self, t, residues=None, cutoff=0.7, proteins=None, atom_names=[], grouped=True, mdtraj=False, leightweight=False):
        """Compute contacts between proteins and given lipids in the system.

        Parameters
        ----------
        t : MDTraj.Trajectory
        residues : list
            Explicitly specify the residues for contact calculations. Really important to speed
            up calculations when you are only interested in a small part of the protein.
            It should be 1-based, meaning the residue numbers should correspond to residues in the input topology.
        cutoff : float
            The maximum distance that two atoms/beads are considered to be in contact.
        atom_names : list
            Restrict selection of lipids only to the following atoms/beads.
        grouped : bool, default=True
            Group lipids in the system according to their headgroup type. E.g. POPC and DOPC
            would be grouped into PC lipids, etc.
        mdtraj : bool, default=False
            Store MDTraj calculated contacts.

        Returns
        -------
        prolint_contacts : dict
            Dictionary of ProLint contacts. Contacts for each residue of each protein
            are stored using the ProLint.LPContacts class.

        """

        if len(atom_names) == 0:
            haystack_indices = self.lipids.l_indices
            lipid_names = self.lipids.lipid_names()
        else:
            df = self.lipids.dataframe
            haystack_indices = list(df[df.name.isin(atom_names)].index)
            lipid_names = df[df.name.isin(atom_names)].resName.unique()

        if grouped:
            PLASMA_LIPIDS = martini_lipids(lipid_names)
        else:
            PLASMA_LIPIDS = {}
            for lip in lipid_names:
                PLASMA_LIPIDS[lip] = [lip]

        if proteins is None:
            proteins = self.proteins

        prolint_contacts = defaultdict(dict)
        for protein in proteins:
            for pc in np.arange(protein.counter):
                print ("Working on protein copy: {}".format(pc))

                if residues is None:
                    residues = protein.residues

                residue_indices = protein.get_indices(residues=residues, df=protein.dataframe[pc])
                per_residue_results = {}
                for chain in range(1, protein.chains+1):

                    for r, idxs in enumerate(residue_indices):
                        # print (residues[r])

                        atoms_count = int(idxs.size / protein.chains)
                        idx = idxs[atoms_count*(chain-1):atoms_count*chain]

                        chain_residue = residues[r] + (protein.n_residues * (chain-1))

                        md_out = md.compute_neighbors(t[1::], cutoff, idx, haystack_indices=haystack_indices)
                        per_residue_results[chain_residue] = LPContacts(t.topology, md_out, PLASMA_LIPIDS,
                                                                self.frames, self.timestep, residue=residues[r])

                prolint_contacts[protein.name][pc] = per_residue_results

        return dict(prolint_contacts)


    def compute_lipid_distances(self, t, protein, distances_dict, PLASMA_LIPIDS, lipids_found, resolution="martini", p=False, grouped=True):

        import itertools

        def get_lipid_atoms(lipid, ldf):
            atoms = ldf[ldf.resName == lipid].name.unique().tolist()
            # getting attoms in this order: P, Ox, Cx, random
            if 'P' in atoms:
                return 'P'
            elif 'O' in ''.join(atoms):
                o = [x for x in atoms if x.startswith('O')]
                o.sort()
                return o[0]
            elif 'C' in ''.join(atoms):
                c = [x for x in atoms if x.startswith('C')]
                c.sort()
                return c[0]
            else:
                return np.random.choice(atoms)

        if p:
            ldf = p.ldf
        else:
            ldf = t.topology.to_dataframe()[0]
        lipid_residue_distances = {}
        lipids = list(distances_dict.keys())

        for lipid in lipids:
            if resolution == "martini":
                headgroup = martini_headgroups(lipid, False)[lipid]
            else:
                headgroup = get_lipid_atoms(lipid, ldf)

            for residue in distances_dict[lipid]:
                residue_distances = {}
                for prot_idx in range(protein.counter):
                    indices = protein.get_indices([residue], copy=prot_idx, suppress=True)[0]

                    df_res = protein.dataframe[prot_idx].loc[indices]
                    if protein.resolution == "martini":
                        residue_atom = "BB"
                    elif protein.resolution == "atomistic":
                        residue_atom = "CA"

                    bb = df_res[df_res.name == residue_atom].index.to_list()
                    if len(bb) == 0:
                        continue

                    lipid_idxs = ldf[
                        (ldf.name == headgroup) &
                        (ldf.resName == lipid)
                    ].index.to_list()

                    d = md.compute_distances(t, list(itertools.product(bb, lipid_idxs)))

                    dist_dict_values = {}
                    dist_sum_list = []
                    dist_label_list = []
                    for dist in range(d.shape[1]):
                        sort_array = d[:, dist].copy()
                        sort_array.sort()

                        dist_sum_list.append(sum(sort_array[-int(len(sort_array)*0.1):]))
                        dist_label_list.append(lipid_idxs[dist])
                        dist_dict_values[lipid_idxs[dist]] = d[:, dist]

                    strongest_binding_label = dist_label_list[np.argmin(dist_sum_list)]
                    residue_distances[prot_idx] = dist_dict_values[strongest_binding_label]

                if lipid_residue_distances.get(lipid):
                    lipid_residue_distances[lipid][residue] = residue_distances
                else:
                    lipid_residue_distances[lipid] = {residue: residue_distances}

        tb = {}
        for k, v in PLASMA_LIPIDS.items():
            if k in lipids_found:
                tightest_lipid = {}
                for vv in v:
                    for resi in lipid_residue_distances[vv].keys():
                        sum_value = []
                        for x in lipid_residue_distances[vv][resi].values():
                            lower_best = np.sort(x)[:int(len(x) * 0.2)]

                            sum_value.append(sum(lower_best))
                        sum_value = sum(sum_value)
                        if tightest_lipid.get(resi):
                            tightest_lipid[resi].append(sum_value)
                        else:
                            tightest_lipid[resi] = [sum_value]

                for kk, sums in tightest_lipid.items():
                    lipid_bound = v[np.argmin(sums)]
                    if tb.get(k):
                        tb[k][kk] = lipid_bound
                    else:
                        tb[k] = {kk: lipid_bound}

        final_results = {}
        for k, v in tb.items():
            for resk, resv in v.items():
                if final_results.get(resv):
                    final_results[resv][resk] = lipid_residue_distances[resv][resk]
                else:
                    final_results[resv] = {resk: lipid_residue_distances[resv][resk]}

        return final_results, dict(time=t.time, protein=[protein.name])


    def compute_distances(self, t, protein, residues, lipid, headgroup, residue_atom=None, distance_co=0.7, percentile_co=0.05, n=None):
        """Compute the distance between a protein residue and lipid atom/bead.

        Parameters
        ----------
        t : MDTraj.Trajectory
            An MDTraj.Trajectory instance.

        protein : ProLint.Protein
            An ProLint.Protein instance.

        residues : int
            The residue number for distance calculation. Numbering starts at 1.

        lipid : str
            The name of the lipid for distance calcualtion.

        headgroup : str
            The name of the atom/bead of the lipid for distance calcualtion. This is usually the headgroup.

        residue_atom : str
            The name of the atom/bead of the residue. This is by default the carbon alpha or backbone bead.
            You can specify any residue atom/bead however.

        distance_co : float
            A cutoff distance that a lipid must satisfy. This cutoff is very useful to ensure not all distances are displayed.
            Only lipids that satisfy this cutoff for percentile_co time are stored.

        percentile_co : float
            The percentage of the trajectory (measured in frames) that a lipid must be within the cutoff for it to be stored. Default is 0.05 (5%) which means
            only lipids that are at least 5% of the time within the default cutoff (0.7) are stored.

        n : dict
            ProLint contact dictionary. If this has already been calculated, it can be supplied to speed up the calculation of
            distances because only lipids that are known to have interacted with the residue are checked. It is optional however.
        """

        import itertools

        whole_dist_dict = {}

        for residue in residues:
            dist_dict = {'Time' : t.time}
            column_count = {}
            min_data = []
            for prot_idx in range(protein.counter):
                indices = protein.get_indices([residue], copy=prot_idx, suppress=True)[0]

                df_res = protein.dataframe[prot_idx].loc[indices]
                if residue_atom is None:
                    if protein.resolution == "martini":
                        residue_atom = "BB"
                    elif protein.resolution == "atomistic":
                        residue_atom = "CA"

                bb = df_res[df_res.name == residue_atom].index.to_list()
                if len(bb) == 0:
                    continue

                ldf = t.topology.to_dataframe()[0]
                if n is None:
                    lipid_idxs = ldf[
                        (ldf.name == headgroup) &
                        (ldf.resName == lipid)
                    ].index.to_list()

                else:
                    lipid_idxs = []
                    lipid_resIDs = n[protein.name][prot_idx][residue].lipids[lipid]
                    if len(lipid_resIDs) == 0:
                        continue
                    for resid in lipid_resIDs:
                        lipid_idx = ldf[
                            (ldf.name == headgroup) &
                            (ldf.resSeq == resid)
                        ].index.to_list()[0]

                        lipid_idxs.append(lipid_idx)

                d = md.compute_distances(t, list(itertools.product(bb, lipid_idxs)))

                column_count[prot_idx] = 0
                for dist in range(d.shape[1]):
                    sort_array = d[:, dist].copy()
                    sort_array.sort()
                    max_allowed = sort_array[int(len(sort_array)*percentile_co)]

                    if max_allowed < distance_co:
                        dist_dict[f'{prot_idx}_{lipid_idxs[dist]}'] = d[:, dist]
                        column_count[prot_idx] += 1
                        min_data.append(max_allowed)

            if len(min_data) == 0:
                min_val = 0
            else:
                min_val = min(min_data)

            meta = dict(
                protein=protein.name,
                lipid=lipid,
                residue=residue,
                headgroup=headgroup,
                residue_atom=residue_atom,
                cutoff=0.7,
                protein_id = list(range(protein.counter)),
                column_count = column_count,
                min_val = min_val
                )

            whole_dist_dict[residue] = (pd.DataFrame(dist_dict), meta)

        return whole_dist_dict



class LPContacts(object):
    """Stores lipid-protein contact information in a readible and usable format.

    Processes the output of the MDTraj.compute_neighbors(), calculating contacts and
    outputing the result in a readable object.

    Attributes
    ----------
    topology : MDTraj.Topology

    mdtraj_contacts : MDTraj.compute_neighbors()

    PLASMA_LIPIDS : dict
        ProLint lipids grouped according to the grouped attribute
        of ProLint.ComputeContacts.compute_neighbors(). E.g. dict(PC=["POPC, "DOPC"]) if grouped=True or
        dict(POPC=["POPC"], DOPC=["DOPC"]) if grouped=False.

    frames : int
        E.g. MDTraj.Trajectory.n_frames

    timestep : int

    residue : int or str
    """
    # Requires import of lipid_species
    def __init__(self, topology, mdtraj_contacts, PLASMA_LIPIDS, frames, timestep, mdtraj=False, residue='residue'):
        """Create LPContacts class and populate attributes"""

        all_lipid_name_id = {}
        occupancy_binaries = {key: np.zeros(frames) for key in PLASMA_LIPIDS.keys()}

        for index, values in enumerate(mdtraj_contacts):

            lipid_ids = {}
            for value in values:

                bead = topology.atom(int(value))
                lipid_name = lipid_species(bead.residue.name, PLASMA_LIPIDS)

                lipid_ids[bead.residue.resSeq] = lipid_name
                occupancy_binaries[lipid_name][index] = 1

            for k, v in lipid_ids.items():
                if all_lipid_name_id.get(v):
                    all_lipid_name_id[v].append(k)
                else:
                    all_lipid_name_id[v] = [k]

        contacts, lipids = {}, {}
        for k in PLASMA_LIPIDS.keys():
            if all_lipid_name_id.get(k):
                contact_list = list((Counter(all_lipid_name_id[k]).values()))
                contacts[k] = [x*timestep for x in contact_list]
                lipids[k] = np.unique(all_lipid_name_id[k])
            else:
                contacts[k] = [0]
                lipids[k] = []

        self.contacts = contacts
        self.occupancy = occupancy_binaries
        self.residue = residue
        self.lipids = lipids

        if mdtraj:
            self.raw_data = mdtraj_contacts


    def __str__(self):
        return "<prolintpy.LPContacts for residue {}>".format(self.residue)


    def __repr__(self):
        return "<prolintpy.LPContacts for residue {}>".format(self.residue)



def retrieve_contacts(n, r, protein=None, contacts='contacts'):
    """Get contact information for a given residue. A similar function is
    retrieve_contacts_flat.

        Parameters
        ----------
        n : ProLint.ComputeContacts
            The dictionary output of ProLint contact calculations.

        r : int
            The residue number (not index)
        protein: str
            If more than one protein in the system, you can specify the protein using
            its name.
        contacts : str
            Type of contact to calculate. 'contacts' or 'occupancy' are allowed.

        Returns
        -------
        contacts : dict
            dictionary of key: value pairs representing lipid: contact * nr_replicates

    """

    if (len(n.keys()) != 1) & (protein is None):
        raise ValueError("""prolintpy.Contacts provided contains information for more than 1 protein. \
Indicate for which protein to calculate contacts using the 'protein' option.""")
    elif protein is None:
        protein = list(n.keys())[0]

    contact_data = defaultdict(list)

    # Get the first key of the input dictionary.
    key1 = list(n.keys())[0]
    key2 = 0
    key3 = list(n[key1][0].keys())[key2]

    is_class = n[key1][key2][key3]

    if isinstance(is_class, LPContacts):
        for pc in n[protein].keys():
            if contacts == 'contacts':
                contact = n[protein][pc][r].contacts
            elif contacts == 'occupancy':
                contact = n[protein][pc][r].occupancy
            else:
                raise ValueError("Only contacts and occupancy are currently implemented.")

            for lipid, values in contact.items():

                contact_data[lipid].append(np.array(values))
    else:
        for pc in n[protein].keys():
            if contacts == 'contacts':
                contact = n[protein][pc][r]
            else:
                raise ValueError("When leightweight is used, only contacts are calcualted (no occupancies, etc.).")

            for lipid, values in contact.items():
                contact_data[lipid].append(np.array(values))

    return contact_data


def retrieve_contacts_flat(n, r, lipid=None, protein=None, contacts='contacts'):
    """Get contact information for a given residue. A similar function is
    retrieve_contacts.

        Parameters
        ----------
        n : ProLint.ComputeContacts
            The dictionary output of ProLint contact calculations.

        r : int
            The residue number (not index)
        lipid : list
            A list of lipids for which to calculate contacts.
        protein: str
            If more than one protein in the system, you can specify the protein using
            its name.
        contacts : str
            Type of contact to calculate. 'contacts' or 'occupancy' are allowed.

        Returns
        -------
        contacts : array
            1D array of all contacts for all protein copies in the system.

    """

    if (len(n.keys()) != 1) & (protein is None):
        raise ValueError("""prolintpy.Contacts provided contains information for more than 1 protein. \
Indicate for which protein to calculate contacts using the 'protein' option.""")
    elif protein is None:
        protein = list(n.keys())[0]

    def is_lipid_defined(c, lipid):
        if (len(c.keys()) != 1) & (lipid is None):
                raise ValueError("""prolintpy.Contacts provided contains information for more than 1 lipid. \n
Indicate for which lipid to calculate contacts using the 'lipid' option.""")
        elif lipid is None:
            lipid = list(c.keys())[0]

        return lipid



    # Get the first key of the input dictionary.
    key1 = list(n.keys())[0]
    key2 = 0
    key3 = list(n[key1][0].keys())[key2]

    is_class = n[key1][key2][key3]

    contact_data = defaultdict(list)
    if isinstance(is_class, LPContacts):
        test_contact = n[protein][0][r].contacts
        lipid = is_lipid_defined(test_contact, lipid)

        for pc in n[protein].keys():
            if contacts == 'contacts':
                contact = n[protein][pc][r].contacts
            elif contacts == 'occupancy':
                contact = n[protein][pc][r].occupancy
            else:
                raise ValueError("Only contacts and occupancy are currently implemented.")

            for l, v in contact.items():
                if l != lipid:
                    continue
                contact_data[l].append(np.array(v))

    else:
        test_contact = n[protein][0][r]

        lipid = is_lipid_defined(test_contact, lipid)

        for pc in n[protein].keys():
            if contacts == 'contacts':
                contact = n[protein][pc][r]
            else:
                raise ValueError("When leightweight is used, only contacts are calcualted (no occupancies, etc.).")

            for l, v in contact.items():
                if l != lipid:
                    continue
                contact_data[l].append(np.array(v))

    return np.concatenate(contact_data[lipid])




def contacts_dataframe(n, p, t, radius, resolution="martini", output_erros=False, co=0, custom_metrics=None, time_unit='us', rt=False, range_type='geo', norm=True):
    """Convert contact information into a DataFrame format. The function calculates
    several contact metrics for each residue of each unique protein in the system.
    Contact information for multiple copies of the same protein are averaged. The following
    are the contacts supported:

    ``mean`` (Mean_Duration) : The average duration of all contacts.

    ``max`` (Longest_Duration) : The longest duration (averaged if more than 1 protein).

    ``sum`` (Sum_of_all_Contacts) : The total sum of all contacts.

    ``lnr`` (Lipid_Number) : The average number of lipids in contact (the total number of contacts
    normalized with respect to the number of frames).

    ``nlnr`` (Normalized_Lipid_Number) : Lipid_Number normalized with respect to the
    number of different lipids (e.g. number of different cholesterols).

    ``occ`` (Occupancy) : For each frame, we give a value of 0 if no lipid of interest is in
    contact with the residue and 1 otherwise. Occupancy is then: sum_of_all_1s / nr_of_frames.

    ``rt``  (Residence_Time) : See here: https://pubs.acs.org/doi/abs/10.1021/ja310577u
    and here: https://www.pnas.org/content/117/14/7803

    Residence time is quite time consuming and by default is not calculated. This can be altered
    using the rt option.

    Parameters
    ----------
    n : ProLint.ComputeContacts
        The dictionary output of ProLint contact calculations.

    p : ProLint.Proteins.system_proteins()
        list containing protein information for the system.

    t : MDTraj.Trajectory

    radius : float
        the cutoff radius used when calculating contacts.

    resolution: string
        the resolution of the data. Default martini.

    output_erros: bool
        Calculate errors and output them in the dataframe? Default is False.
        Currently, errors are only meaningfull if there are multiple proteins in the system.

    co : int
        Discard contacts shorter than co value. Useful when you want to discard contacts
        with a short duration.

    custom_metrics : list
        Provide a custom list of parameters (from the default/supported ones). E.g.
        custom_metrics=['occ'] to calculate only the occupancy.

    time_unit : str
        The timeunit of contact measurements. Either 'us' for microsecond or 'ns' for nanosecond.
        Only used if norm=True.

    rt : bool, default=False
        Calculate the residence time.

    range_type : str
        Residence time can be sped up by supplying a default range. You can specify either a geometric,
        linear, or mixed range ('geo', 'mixed', 'linear'). Fastest and default is 'geo'.

    norm : bool, default=True
        Normalize contacts with respect to time_unit.

    Returns
    -------
    dataframe : pandas.DataFrame
        DataFrame containing contact information for the default metrics.

    """

    RESULTS = defaultdict(list)
    metrics = {
        'mean':'Mean_Duration',
        'max' :'Longest_Duration',
        'sum' :'Sum_of_all_Contacts',
        'lnr' :'Lipid_Number',
        'nlnr':'Normalized_Lipid_Number',
        'occ' :'Occupancy',
    }
    if rt:
        metrics['rt'] = 'Residence_Time'

    if custom_metrics is not None:
        metrics = {key : metrics[key] for key in custom_metrics}

    if resolution == "martini":
        ref_atom = 'BB'
    elif resolution == "atomistic":
        ref_atom = 'CA'

    proteins = list(n.keys())
    for prot_idx, protein in enumerate(proteins):
        residues = list(n[protein][0].keys())

        # Get residue names based on the indices of unique resSeq occurrances.
        # This should work for any ff supported by mdtraj not just cg.
        df = p[prot_idx].dataframe[0]
        resnames = df[df.name == ref_atom].resName.to_numpy()
        selected_resnames = [resnames[x-1] for x in residues]

        for res_idx, residue in enumerate(residues):
            for k, v in metrics.items():
                if k == 'occ':
                    con = retrieve_contacts(n, residue, protein=protein, contacts='occupancy')
                else:
                    con = retrieve_contacts(n, residue, protein=protein)

                if k != 'rt':
                    if (k in ['mean', 'max', 'sum']) & (norm):
                        result = calculate_metric(con, FUNCS[k], co, t, norm=norm, unit=time_unit)
                    else:
                        result = calculate_metric(con, FUNCS[k], co, t)
                else:
                    result = {}
                    for i, j in con.items():
                        sdf = t.topology.to_dataframe()[0]
                        ln = sdf[sdf.resName == i].resSeq.unique().size
                        time = t.time[-1] - t.time[0]
                        result[i] = residence_time(np.concatenate(j), ln, time, range_type=range_type)

                for lipid, values in result.items():
                    RESULTS[v].append(values[0])
                    if output_erros:
                        RESULTS[v + '_Error'].append(values[1])
                    if k == list(metrics.keys())[-1]:
                        RESULTS["Protein"].append(protein)
                        RESULTS["Lipids"].append(lipid)
                        RESULTS["Radius"].append(float(radius))
                        RESULTS["ResID"].append(residue)
                        RESULTS['ResName'].append(selected_resnames[res_idx])

    return pd.DataFrame(RESULTS).sort_values(['Protein', 'Lipids', 'ResID']).reset_index(drop=True)



def retrieve_distances(protein_dataframe, group_lipids, resolution, lipids, top_nr=30):

    if group_lipids and resolution == "martini":
        SYSTEM_LIPIDS = martini_lipids(lipids.lipid_names())
    else:
        SYSTEM_LIPIDS = {}
        for lip in lipids.lipid_names():
            SYSTEM_LIPIDS[lip] = [lip]


    distances_dict = {}
    for protein in protein_dataframe.Protein.unique():
        df = protein_dataframe[protein_dataframe.Protein == protein]
        idx = df.Longest_Duration.sort_values(ascending=False)[:int(top_nr)].index
        df = df[df.index.isin(idx)]
        lipids_found = df.Lipids.unique()
        for lipid in lipids_found:
            if group_lipids:
                for group in SYSTEM_LIPIDS[lipid]:
                    distances_dict[group] = df[df.Lipids == lipid].ResID.to_list()
            else:
                distances_dict[lipid] = df[df.Lipids == lipid].ResID.to_list()

    return distances_dict, SYSTEM_LIPIDS, lipids_found



import yaml
import numpy as np
import pickle

from bokeh.document import Document

from bokeh.plotting import figure
from bokeh.layouts import layout
from bokeh.models import ColumnDataSource
from bokeh.models.widgets import Select
from bokeh.layouts import row
from bokeh.io import show, output_notebook

from bokeh.models import Range1d
from bokeh.themes import Theme


def show_metric_distances(lipid_residue_distances):
    """Visualize the distance between protein residues and lipids.

    """

    output_notebook()

    def distanceApp(doc: Document) -> None:

        distances = lipid_residue_distances[0]
        time = lipid_residue_distances[1]['time']
        proteins = lipid_residue_distances[1]['protein']
        lipids = list(distances.keys())
        residues = list(distances[lipids[0]].keys())
        protein_id = list(distances[lipids[0]][residues[0]].keys())

        residues = [str(x) for x in residues]
        protein_id = [str(x) for x in protein_id]

        source = ColumnDataSource(data=dict(x=[], y=[]))

        protein_name = Select(title="Protein", value=proteins[0], options=proteins, width=100)
        pc = Select(title="Protein Copy", value=protein_id[0], options=protein_id, width=100)
        res = Select(title="Residue Selection", value=residues[0], options=residues, width=100)
        lip = Select(title="Lipid Selection", value=lipids[0], options=lipids, width=100)
        p = figure(plot_width=1500, plot_height=400, tools='pan, box_zoom, ywheel_zoom, save, reset, help')

        color = '#%02x%02x%02x' % tuple(np.random.choice(range(256), size=3))
        p.line(x='x', y='y', line_color=color, source=source, line_width=4)

        p.y_range = Range1d(0, 4)
        p.x_range = Range1d(time[0], time[-1])

        p.toolbar.autohide = True
        p.axis.axis_label_text_font_size = "12pt"
        p.axis.axis_label_text_font_style = "bold"
        p.title.align = 'center'

        p.xaxis.axis_label = "Trajectory Time"
        p.yaxis.axis_label = "Distance (nm)"

        def update():

            protein_selected = protein_name.value
            copy_selected = pc.value
            residue_selected = res.value
            lipid_selected = lip.value

            residues = list(distances[lipid_selected].keys())
            if int(residue_selected) not in residues:
                residue_selected = residues[0]

            res.options = [str(x) for x in residues]

            y = distances[lipid_selected][int(residue_selected)][int(copy_selected)]
            source.data = dict(
                x=time,
                y=y
            )

        controls = [protein_name, pc, res, lip]
        for control in controls:
            control.on_change('value', lambda attr, old, new: update())

        inputs2 = row([*controls])
        layout1 = layout([[inputs2]])
        layout2 = layout([p])

        update()

        doc.add_root(layout1)
        doc.add_root(layout2)

        doc.title = "Distance Calculations"
        doc.theme = Theme(json=yaml.load("""
            attrs:
                Figure:
                    toolbar_location: above
                    height: 500
                    width: 1000
                Grid:
                    grid_line_dash: [6, 4]
                    grid_line_color: black
        """, Loader=yaml.FullLoader))

    return show(distanceApp)




import yaml
import numpy as np

from bokeh.document import Document

from bokeh.plotting import figure
from bokeh.layouts import layout
from bokeh.models import ColumnDataSource
from bokeh.models.widgets import Select
from bokeh.layouts import row
from bokeh.io import show, output_notebook

from bokeh.themes import Theme

from bokeh.models import Range1d


def show_distances(distances, norm_factor=1000000):
    """Visualize the distance between protein residues and lipids.

    Parameters
    ----------
    distances : tuple
        This is the output of a pl.ComputeContacts.compute_distances() call.
        See docs for more help.

    norm_factor : int
        Normalize time dimension (i.e. convert time to nano or microsecond).
        Default is 1000000 and will convert picoseconds to microseconds.

    """

    output_notebook()

    def distanceApp(doc):

        global df

        residues = list(distances.keys())
        df, meta = distances[residues[0]]
        residues = [str(x) for x in residues]

        time = df['Time'].to_list()

        proteins = [meta['protein']]
        protein_id = meta['protein_id']
        protein_id = [str(x) for x in protein_id]
        min_val = meta['min_val']

        columns = []
        for resid in residues:
            for col in distances[int(resid)][0].columns[1:]:
                columns.append(col)

        source_dict = {}
        for i in np.unique(columns):
            source_dict[i] = []
        source_dict['x'] = []

        source = ColumnDataSource(data=source_dict)

        protein_name = Select(title="Protein", value=proteins[0], options=proteins, width=100)
        pc = Select(title="protein_id", value=protein_id[0], options=protein_id, width=100)
        res = Select(title="Residue Selection", value=residues[0], options=residues, width=100)

        p = figure(plot_width=1500, plot_height=400, tools='pan, box_zoom, ywheel_zoom, save, reset, help')

        for i in np.unique(columns):
            color = '#%02x%02x%02x' % tuple(np.random.choice(range(256), size=3))
            p.line(x='x', y=i, line_color=color, source=source, line_width=2)

        p.y_range = Range1d(min_val-(min_val*0.15), 2.5)
        p.x_range = Range1d(time[0], time[-1])

        p.toolbar.autohide = True
        p.axis.axis_label_text_font_size = "12pt"
        p.axis.axis_label_text_font_style = "bold"
        p.title.align = 'center'

        p.xaxis.axis_label = "Trajectory Time"
        p.yaxis.axis_label = "Distance (nm)"


        def update():

            df = distances[int(res.value)][0]
            time = df['Time'].to_list()
            df['NaNs'] = ['NaN'] * len(time)

            show_columns = [x for x in df.columns if x.startswith(pc.value + '_')]
            not_columns = [x for x in np.unique(columns) if x not in show_columns]

            new_source_dict = {}
            new_source_dict['x'] = df['Time']
            for col in show_columns:
                new_source_dict[col] = df[col]

            for col in not_columns:
                new_source_dict[col] = df['NaNs']

            source.data = new_source_dict


        controls = [protein_name, pc, res]
        for control in controls:
            control.on_change('value', lambda attr, old, new: update())

        inputs2 = row([*controls])
        layout1 = layout([[inputs2]])
        layout2 = layout([p])

        update()

        doc.add_root(layout1)
        doc.add_root(layout2)

        doc.title = "Distance Calculations"
        doc.theme = Theme(json=yaml.load("""
            attrs:
                Figure:
                    toolbar_location: above
                    height: 500
                    width: 800
                Grid:
                    grid_line_dash: [6, 4]
                    grid_line_color: black
        """, Loader=yaml.FullLoader))

    return show(distanceApp)



import yaml
import numpy as np
import pandas as pd
import colorcet as cc

from bokeh.document import Document

from bokeh.plotting import figure
from bokeh.layouts import layout
from bokeh.models import ColumnDataSource
from bokeh.models.widgets import Slider, Select, TextInput
from bokeh.layouts import row
from bokeh.palettes import Reds
from bokeh.transform import linear_cmap
from bokeh.io import show, output_notebook

from bokeh.themes import Theme

def show_points(df=None, filename=None, **kwargs):
    """Visualize lipid-protein contacts using a Bokeh scatter plot.

    Parameters
    ----------

    df : pandas.DataFrame
        A contacts dataframe. This will be the output of pl.contacts_dataframe() command.

    filename: string
        A string pointing to the location of a csv file containing data that can be loaded into a pandas DataFrame.

    size : int
        The size of points to dispaly. Default is 10.

    """
    if df is None and not filename:
        raise Exception("show_points can take either a pandas DataFrame or a csv file that can be loaded into a pandas DataFrame.")
    elif filename:
        df = pd.read_csv(filename)

    output_notebook()

    def pointsApp(doc):

        radii = [str(x) for x in df.Radius.unique()]

        # Create Input controls
        number = Slider(title="Value Cutoff", value=0, start=0, end=4, step=0.1, width=150)

        residue = TextInput(title="Residue name (3 letter code):", width=200)

        gpcr = Select(title="Proteins", value=list(df.Protein.unique())[0],
                    options=list(df.Protein.unique()), width=100)

        lipid = Select(title="Lipids", value=list(df.Lipids.unique())[0],
                    options=list(df.Lipids.unique()), width=100)

        radius = Select(title="Radius", value=radii[-1], options=radii, width=100)

        options = list(df.columns)[:-5] + ['ResID']
        x_axis = Select(title="X Axis", options=options, value="ResID", width=150)
        y_axis = Select(title="Y Axis", options=options, value=options[0], width=150)


        cc_colors = [x for x in cc.all_original_names() if x.startswith('linear') or x.startswith('rainbow')]
        cmap = Select(title="Colormap", options=cc_colors, value='linear_kryw_0_100_c71', width=150)

        # Create Column Data Source that will be used by the plot
        source = ColumnDataSource(data=dict(x=[], y=[], ResName=[], ResID=[], Protein=[]))

        TOOLTIPS=[
            ("ResName", "@ResName"),
            ("ResID", "@ResID"),
            ("Value", "@y")
        ]

        mapper = linear_cmap(field_name='y', palette=cc.CET_L19,
                            low=df[df.Protein == gpcr.value][y_axis.value].min(),
                            high=df[df.Protein == gpcr.value][y_axis.value].max())

        p = figure(plot_width=1100, plot_height=600, tooltips=TOOLTIPS,)

        global c
        c = p.circle(x="x", y="y", source=source, line_color='black', fill_color=mapper, **kwargs)

        p.toolbar.autohide = True
        p.axis.axis_label_text_font_size = "12pt"
        p.axis.axis_label_text_font_style = "bold"
        p.title.align = 'center'

        def update(df):
            y_value = y_axis.value
            x_value = x_axis.value

            df = df[
                (df[y_value] >= number.value) &
                (df['Protein'] == gpcr.value) &
                (df['Lipids'] == lipid.value) &
                (df['Radius'] == float(radius.value))
            ]
            if (residue.value != ""):
                df = df[df.ResName.str.contains(residue.value.upper())==True]

            mapper = linear_cmap(field_name='y', palette=cc.palette[cmap.value],
                                low=df[df.Protein == gpcr.value][y_value].min(),
                                high=df[df.Protein == gpcr.value][y_value].max())

            c.glyph.fill_color = mapper

            p.xaxis.axis_label = x_value
            p.yaxis.axis_label = y_value
            p.title.text = "Showing %d Data Points  " % len(df)

            source.data = dict(
                x=df[x_value],
                y=df[y_value],
                ResName=df["ResName"],
                ResID=df["ResID"],
                Protein=df["Protein"],
            )

        controls = [number, gpcr, lipid, radius, y_axis, x_axis, residue, cmap]
        for control in controls:
            control.on_change('value', lambda attr, old, new: update(df))

        inputs = row(*controls)
        inputs2 = row([gpcr, lipid, radius, residue])
        inputs3 = row([number, x_axis, y_axis, cmap])

        layout1 = layout([[inputs2]])
        layout2 = layout([p])
        layout3 = layout([inputs3])

        update(df)

        doc.add_root(layout1)
        doc.add_root(layout2)
        doc.add_root(layout3)
        doc.title = "Scatter Application"

    return show(pointsApp)



import os
import pickle
import numpy as np

import matplotlib as mpl
import matplotlib.cm as cm

# bokeh aplication
from bokeh.document import Document

from bokeh.plotting import figure
from bokeh.layouts import layout, column, row
from bokeh.models import ColumnDataSource, LinearColorMapper, ColorBar, BasicTickFormatter
from bokeh.models.widgets import Slider, Select, RangeSlider

from bokeh.io import show, output_notebook

def flat_density(t, lipid, frames):

    if len(lipid) == 1:
        selection_string = "resname {}".format(lipid[0])
    else:

        s_st_list = ["{} or resname ".format(x) for x in lipid[:-1]]

        selection_string = "resname "
        for s in s_st_list:
            selection_string += s
        selection_string += lipid[-1]

    lipid_ndx = t.topology.select(selection_string)
    lipid_xyz = t.xyz[:, lipid_ndx, :]

    def slice_array(arr, slice_by):
        mask = np.ones_like(arr, dtype=bool)
        mask[::int(slice_by)] = False
        xshape = int(frames - mask[::int(slice_by)].shape[0])
        arr = arr[mask].reshape(xshape*arr.shape[1], 3)

        return arr

    lipid_xyz = lipid_xyz.reshape(lipid_xyz.shape[0]*lipid_xyz.shape[1], 3)

    array_memory_size = len(pickle.dumps(lipid_xyz)) / 1024
    while array_memory_size > 5000:
        lipid_xyz = slice_array(lipid_xyz, 10)

        array_memory_size = len(pickle.dumps(lipid_xyz)) / 1024

    return lipid_xyz


def show_density(t, lipids, resolution, filenames=None):
    """Visualize the preferential localization of lipids using 2D density maps.

    Parameters
    ----------

    t : MDTraj.Trajectory

    lipids: list
        List of lipids to visualize their density. Currently, only single lipids are supported. No grouping.

    resolution: string
        Either "martini" or "atomistic"

    filename: list, Optional
        A list of strings pointing to the location of npy files that are saved by ProLint.

    """

    if filenames is not None:
        lipids = []
        for filename in filenames:
            if filename.split('/')[-1].startswith('prot'):
                prot_xyz = np.load(filename)
            else:
                lipids.append(filename)
        l_xyz = np.load(lipids[0])
    else:
        from prolintpy.core.systemtopology import Proteins
        p = Proteins(t.topology, resolution=resolution)
        prot_xyz = t.xyz[0, p.p_indices, :]
        l_xyz = flat_density(t, lipids, t.n_frames)

    output_notebook()

    def densityApp(doc):
        """
        Handler function for the density application.
        """
        nonlocal t, lipids, resolution, prot_xyz, l_xyz

        all_mpl_cmaps = [
                    'viridis', 'plasma', 'inferno', 'cividis',
                    'Greys', 'Purples', 'Blues', 'Greens', 'Oranges', 'Reds',
                    'RdYlBu', 'RdYlGn', 'Spectral','Spectral_r', 'coolwarm', 'coolwarm_r', 'seismic',
                    'YlOrBr', 'YlOrRd', 'OrRd', 'PuRd', 'RdPu', 'BuPu',
                    'GnBu', 'PuBu', 'YlGnBu', 'PuBuGn', 'BuGn', 'YlGn',
                    'PiYG', 'PRGn', 'BrBG', 'PuOr', 'RdGy', 'RdBu',
        ]
        lipids = [x for x in lipids if not x.startswith('prot')]

        x_min, x_max = l_xyz[:, 0].min(), l_xyz[:, 0].max()
        y_min, y_max = l_xyz[:, 1].min(), l_xyz[:, 1].max()
        z_min, z_max = l_xyz[:, 2].min(), l_xyz[:, 2].max()

        # widgets
        color_map = Select(title="Colormap", value="viridis",
                        options=all_mpl_cmaps, width=120)

        lipid = Select(title="Lipids", value=lipids[0],
                    options=lipids, width=100)

        number = Slider(title="Number of Bins", value=150, start=80, end=500, step=10, width=200)

        protein = Select(title="Show Protein", value="No",
                        options=["Yes", "No"], width=90)

        zrange = RangeSlider(start=z_min, end=z_max,
                            value=(z_min, z_max), step=0.2, title="Z-Axis Range",
                            callback_policy='mouseup', width=352)

        cbar_range = Slider(value=256, start=0, end=256, step=1, height=600,
                            orientation='vertical', callback_policy='mouseup', margin=(15, 0),
                            tooltips=False, show_value=False)

        # colorschemes
        colormap =cm.get_cmap(color_map.value)
        bokehpalette = [mpl.colors.rgb2hex(m) for m in colormap(np.arange(colormap.N))]

        def lipid_density(array, bins=160):
            """Given a 2D array, containing the x, y values of a lipid,
            return its histogram with edges."""
            x = array[:, 0]
            y = array[:, 1]

            lipid, e1, e2 = np.histogram2d(x, y, density=True, bins=bins)

            return lipid, e1, e2

        # Plot histogram image
        H, xe, ye = lipid_density(l_xyz, number.value)
        minx = np.abs(xe.min())
        miny = np.abs(ye.min())

        p1 = figure(plot_height=640, plot_width=640,  tools='')
        image_source = ColumnDataSource(data=dict(image=[]))
        img = p1.image(image="image", x=xe[0], y=ye[0], dw=xe[-1] + minx, dh=ye[-1] + miny, palette=bokehpalette, source=image_source)

        circle_prot_source = ColumnDataSource(data=dict(x1=prot_xyz[:, 0], y1=prot_xyz[:, 1]))
        p1.circle(x="x1", y="y1", source=circle_prot_source, size=2, fill_alpha=0.2)

        cb_palette = LinearColorMapper(palette=bokehpalette, low=H.min(), high=H.max())
        color_bar = ColorBar(color_mapper=cb_palette, width=8,  location=(0,0), label_standoff=10)
        color_bar.formatter = BasicTickFormatter(use_scientific=False)
        p1.add_layout(color_bar, 'right')

        # Make graph pretty
        p1.xgrid.grid_line_color = None
        p1.ygrid.grid_line_color = None
        p1.xaxis.major_tick_line_color = None
        p1.xaxis.minor_tick_line_color = None
        p1.yaxis.major_tick_line_color = None
        p1.yaxis.minor_tick_line_color = None
        p1.xaxis.major_label_text_font_size = '0pt'
        p1.yaxis.major_label_text_font_size = '0pt'
        p1.grid.visible = False
        p1.toolbar.logo = None
        p1.toolbar_location = None


        def update_all(cond=False, cmap=False):
            """
            Update the image showing all proteins.
            """
            if cond:
                # For efficiency execute only if GPCR structure changes
                if filenames is not None:
                    update_l_xyz = np.load(lipid.value)
                else:
                    update_l_xyz = flat_density(t, [lipid.value], t.n_frames)

                x_min, x_max = update_l_xyz[:, 0].min(), update_l_xyz[:, 0].max()
                y_min, y_max = update_l_xyz[:, 1].min(), update_l_xyz[:, 1].max()
                z_min, z_max = update_l_xyz[:, 2].min(), update_l_xyz[:, 2].max()
                zrange.start = z_min
                zrange.end   = z_max

                index = np.where((update_l_xyz[:, 2] > zrange.value[0]) &
                                (update_l_xyz[:, 2] < zrange.value[1]))
                l_xyz_new = update_l_xyz[index]
            else:
                l_xyz_new = l_xyz

            if cmap:
                # For efficiency execute only if image colormap changes
                cb_cut_value = 256 - cbar_range.value

                cmap = color_map.value
                colormap = cm.get_cmap(cmap)
                bokehpalette = [mpl.colors.rgb2hex(m) for m in colormap(np.arange(colormap.N))]
                bp_i = 0
                while bp_i < len(bokehpalette[:cb_cut_value]):
                    bokehpalette[bp_i] = '#ffffff'
                    bp_i += 1

                img.glyph.color_mapper.palette = bokehpalette
                color_bar.color_mapper.palette = bokehpalette

            # Update histogram image
            H, xe, ye = lipid_density(l_xyz_new, number.value)
            minx = np.abs(xe.min())
            miny = np.abs(ye.min())

            img.glyph.dw = xe[-1] + minx
            img.glyph.dh = ye[-1] + miny

            # update image source
            image_source.data = dict(image=[H])

        def update_protein():
            """
            Update the protein representation.
            """

            if protein.value == "Yes":
                circle_prot_source.data = dict(
                    x1=prot_xyz[:, 0],
                    y1=prot_xyz[:, 1]
                )

            elif protein.value == "No":
                circle_prot_source.data = dict(
                    x1=[],
                    y1=[]
                )

        def update_cbar():

            cb_cut_value = 256-cbar_range.value

            cmap = color_map.value
            colormap = cm.get_cmap(cmap)
            bokehpalette = [mpl.colors.rgb2hex(m) for m in colormap(np.arange(colormap.N))]

            bp_i = 0
            while bp_i < len(bokehpalette[:cb_cut_value]):
                bokehpalette[bp_i] = '#ffffff'
                bp_i += 1

            img.glyph.color_mapper.palette = bokehpalette
            color_bar.color_mapper.palette = bokehpalette


        # event listeners
        controls = [lipid, zrange]
        for control in controls:
            control.on_change('value', lambda attr, old, new: update_denstype(cond=True))

        number.on_change('value', lambda attr, old, new: update_denstype())
        color_map.on_change('value', lambda attr, old, new: update_denstype(cmap=True))
        protein.on_change('value', lambda attr, old, new: update_protein())
        cbar_range.on_change('value', lambda attr, old, new: update_cbar())

        # deal with what gets updated and what not.
        def update_denstype(cond=False, cmap=False):
            update_all(cond, cmap)
            update_protein()

        update_denstype()
        input1 = row([lipid, color_map, protein])
        input2 = row([p1, cbar_range])
        input3 = row([number, zrange])
        input3 = column([input1, input2, input3])

        l = layout([input3])

        doc.add_root(l)
        doc.title = "Density App"

    return show(densityApp)



from setuptools import setup, find_packages

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name='prolintpy',
    version='0.9.6',
    description='Automated analyis and visualization of lipid-protein interactions.',
    url="https://github.com/ProLint/prolintpy",
    author='Besian I. Sejdiu',
    author_email="besian.sejdiu@stjude.org",
    license='MIT',
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires='>=3.6',
    packages=find_packages(),
    install_requires=[
	'numpy', 
	'pandas==0.24.0', 
	'scipy', 
	'mdtraj==1.9.5', 
	'pyyaml', 
	'colorcet', 
	'bokeh==1.4.0', 
	'networkx', 
	'jupyterlab', 
	'nglview', 
	'matplotlib'],
)

